import java.util.ArrayList;
import java.util.Map;


/**
 * @author Karl Janson
 * @since 11/23/14
 * Class to analyze lines taken from assembler file
 */
public class AsmAnalyzer {

    /**
     * Mehtod to parse numbers in any format
     * and convert them into binary numbers.
     * @param operand Instruction's operand
     * @return Parsed operand as an binary number string
     * @throws AsmException assembler exception
     */
    private String intParser(String operand, int bitWidth) throws AsmException {

        boolean negative = false;
        int intOperand;

        //Is the operand negative? If so, set the negative flag and remove the '-' sign for parsing
        if (operand.startsWith("-")) {
            negative = true;
            operand = operand.substring(1, operand.length());
        }

        //Is the operand a hexadecimal number?
        if (operand.contains("0X") || operand.contains("0x")) {
            //Cut out the '0x' part and convert to int
            intOperand = Integer.parseInt(operand.substring(2, operand.length()), 16);
        }
        //Or is it binary?
        else if (operand.contains("0B") || operand.contains("0b")) {
            //Cut out the '0b' part
            operand = operand.substring(2, operand.length());

            //Is it a negative binary number?
            if ((operand.length() == bitWidth) && (operand.startsWith("1"))) {
                    return String.format("%" + bitWidth +"s", operand).replace(' ', '1');
            }
            intOperand = Integer.parseInt(operand.substring(0, operand.length()), 2);
        } else {
            intOperand = Integer.parseInt(operand);
        }

        //Is it a negative number?
        if (negative) {
            intOperand *= -1;
        }


        //Check for overflow
        if (intOperand < ((Math.pow(2,bitWidth)) / 2) * -1  || intOperand > ((Math.pow(2,bitWidth)) / 2) - 1) {
            throw new AsmException("Memory overflow. The number is out of scope of " + bitWidth + "-bit signed integers:\n");
        }

        //Convert the operand back to a binary string
        operand = Integer.toBinaryString(intOperand);
        if (operand.length() < bitWidth) {
            operand = String.format("%" + bitWidth +"s", operand).replace(' ', '0');
        } else {
            operand = operand.substring(operand.length() - bitWidth, operand.length());
        }
        return operand;
    }

    /**
     * Method to analyze the operand.
     * @param operand The instruction's operand
     * @param operandType Specified the operand's type
     * @param labelList List of known labels
     * @param constList List of known constants
     * @return Returns string of analyzed binary integer
     * @throws AsmException
     */
    private String operandAnalyzer(String operand, String operandType,
                                   Map<String, String> labelList,
                                   Map<String, String> constList,
                                   int bitWidth) throws AsmException {

        //No operand
        if (operandType == null) {
            return String.format("%" + bitWidth +"s", '0').replace(' ', '0');
        }

        /*Operand is a number*/
        else if (operandType.equals("ICODE")) {
            return intParser(operand, bitWidth);
        }

        /*Operand is a constant*/
        else if (operandType.equals("MEM_CONST")) {
            operand = constList.get(operand);
            if (operand == null) {
                throw new AsmException("Unknown constant:\n");

            } else {
                return intParser(operand, bitWidth);
            }
        }

        /*Operand is a label*/
        else if (operandType.equals("LABEL")) {
            //Reformats  and gets the operand
            operand = labelList.get(operand.substring(1, operand.length()) + ":");
            if (operand == null) {
                throw new AsmException("Unknown constant:\n");

            } else {
                return intParser(operand, bitWidth);
            }
        /*Operand is of and unknown type*/
        } else {
            throw new AsmException("Unknown operand type:\n");
        }
    }

    /**
     * Method to analyze the instruction and turn it into machine code.
     * @param instructionMap map of instructions
     * @param labelList List of known labels
     * @param constList List of known constants
     * @param fileBuf File buffer
     * @return Returns array of machine code strings
     */
    public ArrayList<String> analyze(Map<String, String> instructionMap,
                                     Map<String, String> labelList,
                                     Map<String, String> constList,
                                     ArrayList<String[]> fileBuf,
                                     int bitWidth) {

        ArrayList<String> machineCode = new ArrayList<String>();
        String opcode;
        String operand;
        String operandType;
        String lineNr;

        for (String[] instructionLine : fileBuf) {
            opcode = instructionLine[0];
            operand = instructionLine[1];
            operandType = instructionLine[2];
            lineNr = instructionLine[3];

            if (opcode == null) {
                System.err.println("Illegal opcode on line " + lineNr);
            } else {
                opcode = instructionMap.get(opcode);
                if (opcode == null) {
                    System.err.println("Error on line " + lineNr
                            + "\nOpcode " + instructionLine[0] + "not found");
                }
            }
            try {
                operand = operandAnalyzer(operand, operandType, labelList, constList, bitWidth);
            } catch (AsmException e){
                System.err.println("Error at line " + lineNr + "\n" + e.getMessage() + operand);
                return null;
            }
            machineCode.add(String.format("%16s", opcode + operand).replace(' ', '0'));
        }
        return machineCode;
    }
}
