import org.antlr.v4.runtime.misc.NotNull;

import java.util.ArrayList;
import java.util.Map;
import java.util.TreeMap;

/**
 * Loader class for assembler file.
 * @author Karl Janson
 * @since 11/23/14
 */
public class AsmLoader extends AssemblerBaseListener {

    private int address = 0;
    private boolean exceptionOccurred = false;

    private Map<String, String> labelList = new TreeMap<String, String>(String.CASE_INSENSITIVE_ORDER);
    private Map<String, String> constList = new TreeMap<String, String>(String.CASE_INSENSITIVE_ORDER);
    private ArrayList<String[]> fileBuf = new ArrayList<String[]>();

    @Override
    public void exitInstruction(@NotNull AssemblerParser.InstructionContext ctx) {
        super.exitInstruction(ctx);

        //Build the instruction data to save to the buffer
        String[] linebuf = new String[5];
        linebuf[0] = null; //Opcode
        linebuf[1] = null; //Operand
        linebuf[2] = null; //Operand type
        linebuf[3] = Integer.toString(ctx.getStart().getLine()); //Line number

        //Detect the type of the operand
        for (int i = 0; i < ctx.getChildCount(); i++) {
            if (ctx.getChild(i) instanceof AssemblerParser.InameContext) {
                linebuf[0] = ctx.getChild(i).getChild(0).getText();
            } else if (ctx.getChild(i) instanceof AssemblerParser.IcodeContext) {
                linebuf[1] = ctx.getChild(i).getChild(0).getText();
                linebuf[2] = "ICODE";
            } else if (ctx.getChild(i) instanceof AssemblerParser.LabelContext) {
                linebuf[1] = ctx.getChild(i).getChild(0).getText();
                linebuf[2] = "LABEL";
            } else if (ctx.getChild(i) instanceof AssemblerParser.Mem_constContext) {
                linebuf[1] = ctx.getChild(i).getChild(0).getText();
                linebuf[2] = "MEM_CONST";
            }
        }
        fileBuf.add(linebuf);
        address++;
    }

    @Override
    public void exitMem_assign(@NotNull AssemblerParser.Mem_assignContext ctx) {
        super.exitMem_assign(ctx);
            if (constList.get(ctx.MEM_CONST().getText()) == null) {
                constList.put(ctx.MEM_CONST().getText(), ctx.ICODE().getText());
            } else {
                System.err.println("Error on line " + ctx.getStart().getLine()
                        + ":\nConstant " + ctx.MEM_CONST().getText() + " already exists");
                exceptionOccurred = true;
            }
    }

    @Override
    public void exitLabel_decl(@NotNull AssemblerParser.Label_declContext ctx) {
        super.exitLabel_decl(ctx);
            if (labelList.get(ctx.LABEL_DECL().getText()) == null) {
                labelList.put(ctx.LABEL_DECL().getText(), "0b" + Integer.toBinaryString(address));
            } else {
                System.err.println("Error on line " + ctx.getStart().getLine()
                        + ":\nLabel " + ctx.LABEL_DECL().getText() + " already exists");
                exceptionOccurred = true;
            }
    }

    public boolean getExceptionOccurred() {
        return exceptionOccurred;
    }

    public Map<String,String> getLabelList() {
        return labelList;
    }
    public Map<String,String> getConstList() {
        return constList;
    }
    public ArrayList<String[]> getFileBuf() {
        return fileBuf;
    }
}
