import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

import java.io.*;
import java.util.ArrayList;
import java.util.Map;

/**
 * @author Karl Janson
 * @since 11/22/14
 */
public class Main {

    /*Declare commandline options*/

    @Option(name = "-h", aliases = { "--help" }, usage = "Display this message")
    private boolean helpDefined;

    @Option(name = "-no", aliases = { "--no-file" },
            usage = "Do not save output vectors to file, only display them")
    private boolean noOutFile;

    @Option(name = "-o", usage = "Path to output file "
            + "(optional, default machinecode.txt)", metaVar = "OUTFILE")
    private String outFile = "";

    @Option(name = "-i", usage = "Path to instructions file "
            + "(optional, default intructions.txt)", metaVar = "INSTRFILE")
    private String instFile = "";

    @Option(name = "-a", usage = "Path to assembler file ", metaVar = "ASMFILE")
    private String asmFile = "";

    @Option(name = "-dI", usage = "Debug mode: show syntax tree for instructions file")
    private boolean showInstrTree;

    @Option(name = "-dA", usage = "Debug mode: show syntax tree for assembler file")
    private boolean showAsmTree;

    @Option(name = "-bw", aliases = { "--bitwidth" }, usage = "Specify the bitwidth (defaults to 8)")
    private int bitWidth = 8;


    /**
     * Actual main function. Just used as a wrapper for doMain().
     * @param args System arguments
     * @throws Exception
     */
    public static void main(String[] args) throws Exception {
        new Main().doMain(args);
    }


    /**
     * Main function.
     * @param args System arguments from Main()
     * @throws Exception
     */
    public void doMain(String[] args) throws Exception {

        /*Process command line arguments*/
        CmdLineParser cmdParser = new CmdLineParser(this);
        cmdParser.setUsageWidth(80);

        try {
            // parse the arguments.
            cmdParser.parseArgument(args);

            if (helpDefined) {
                //Just throw an empty error message to print the help screen and return
                throw new CmdLineException(cmdParser, "");
            }

            if (showAsmTree && showInstrTree) {
                throw new CmdLineException(cmdParser,
                        "Error: Only one of the debug mode switches can be used at the time\n");
            }

            if (showInstrTree) {
                System.out.println("Debug mode: will show instruction file syntax tree");
            }

            if (showAsmTree) {
                System.out.println("Debug mode: will show assembler file syntax tree");
            }

            if (noOutFile) {
                System.out.println("Will not save output vectors to file");
            } else {
                if (outFile.equals("")) {
                    System.out.println("Output file not specified, using default: machinecode.txt");
                    outFile = "machinecode.txt";
                }
            }

            if (asmFile.equals("")) {
                throw new CmdLineException(cmdParser, "Error: Input file not specified\n");

            }

            if (instFile.equals("")) {
                System.out.println("Instructions file not specified, using default: intructions.txt");
                instFile = "intructions.txt";

            }

        } catch (CmdLineException e) {
            // if there's a problem in the command line,
            // you'll get this exception. this will report
            // an error message.
            System.err.println(e.getMessage());
            System.err.println("Usage: picoAssembler [options] -a ASMFILE");
            // print the list of available options
            cmdParser.printUsage(System.err);
            System.err.println();
            System.err.println("Only one debug mode switch can be used at the time,"
                    + " in debug mode no real work is done");

            return;
        }

        /*Read and process the instruction file*/
        InputStream instrStream = System.in;
        if (instFile != null) {
            try {
                instrStream = new FileInputStream(instFile);
            } catch (IOException e) {
                System.err.println("Error: Can not read instruction file " + instFile);
                return;
            }
        }

        Map<String, String> intructionMap;

        InstrReader instrReader = new InstrReader();
        intructionMap = instrReader.readInstr(instrStream, showInstrTree);

        if (intructionMap == null) {
            if (showInstrTree) {
                return;
            } else {
                System.err.println("picoAssembler terminated because of an "
                        + "error while reading the instructions file");
                return;
            }
        }

        /*Read and process the assembler file*/
        InputStream asmStream = System.in;
        if (asmFile != null) {
            try {
                asmStream = new FileInputStream(asmFile);
            } catch (IOException e) {
                System.err.println("Error: Can not read assembler file " + asmFile);
                return;
            }
        }

        AsmReader asmReader = new AsmReader();
        AsmLoader loader = asmReader.readAsm(asmStream, showAsmTree);

        if (loader == null) {
            if (showAsmTree) {
                return;
            } else {
                System.err.println("picoAssembler terminated because of an "
                        + "error while reading the assembler file");
                return;
            }
        }

        AsmAnalyzer analyzer = new AsmAnalyzer();
        ArrayList<String> machineCode = analyzer.analyze(intructionMap,
                loader.getLabelList(),
                loader.getConstList(),
                loader.getFileBuf(),
                bitWidth);

        if (machineCode == null) {
                System.err.println("picoAssempler terminated because of an "
                        + "error while processing the assembler code");
                return;
        }

        System.out.println("\nAssembled machine code:\n");
        String outputBuf = "";

        for (String line : machineCode) {
            if (!noOutFile) {
                outputBuf += line + "\n";
            }
            System.out.println(line);
        }

        if (!noOutFile) {
            try {

                File file = new File(outFile);

                if (!file.exists()) {
                    file.createNewFile();
                }

                FileWriter fw = new FileWriter(file.getAbsoluteFile());
                BufferedWriter bw = new BufferedWriter(fw);
                bw.write(outputBuf);
                bw.close();

            } catch (IOException e) {
                System.err.format("Error writing to output file " + outFile);
            }
            System.out.println("\nMachine code saved to " + outFile);
        }

        System.out.println("\nAll done, thank you!");
    }
}

